__program_name__ = "Curare (Customizable and Reproducible Analysis Pipeline for RNA-Seq Experiments)"
__version__ = "0.4.3"
__author__ = "Patrick Blumenkamp"
__author_email__ = "patrick.blumenkamp@computational.bio.uni-giessen.de"
